from collections.abc import Generator
import inspect
import pytest
import re
import tree_sitter_talon
import typing


Context: typing.TypeAlias = typing.Union[None, typing.Mapping[str, typing.Any]]


def short(sig: inspect.Signature) -> str:
    out = str(sig)
    out = out.replace("tree_sitter_type_provider.node_types.", "")
    out = re.sub(r"ForwardRef\('(\w+)'\)", r"\1", out)
    out = out.replace("typing.", "")
    out = out.replace("NoneType", "None")
    out = out.replace("abc.", "")
    out = out.replace("~Result", "Result")
    return out


def function_signatures(
    object: object, ctx: Context = None
) -> Generator[str, None, None]:
    for name, fun in inspect.getmembers(object, inspect.isfunction):
        if not (name.startswith("_") or name in ["to_dict", "to_json"]):
            try:
                funsig = inspect.signature(fun, globals=ctx, eval_str=True)
                yield f"{name}{short(funsig)}"
            except ValueError:
                pass


def class_signatures(object: object, ctx: Context = None) -> Generator[str, None, None]:
    for name, cls in inspect.getmembers(object, inspect.isclass):
        if not name.startswith("_"):
            try:
                clssig = inspect.signature(cls, globals=ctx, eval_str=True)
                yield f"{name}{short(clssig)}"
                for funsigstr in function_signatures(cls, ctx=ctx):
                    yield f"  {funsigstr}"
            except ValueError:
                pass


@pytest.mark.golden_test("api.yml")
def test_talon_api(golden):
    assert golden["input"] is None

    ctx = globals() | tree_sitter_talon.__dict__

    output: list[str] = []
    output.extend(function_signatures(tree_sitter_talon.__class__, ctx=ctx))
    output.extend(function_signatures(tree_sitter_talon, ctx=ctx))
    output.extend(class_signatures(tree_sitter_talon.__class__, ctx=ctx))
    output.extend(class_signatures(tree_sitter_talon, ctx=ctx))

    assert "\n".join(output) == golden.out["output"]



from pathlib import Path
from pkg_resources import resource_filename  # type: ignore
from tree_sitter_type_provider import *

import sys
import tree_sitter
import typing
import os


class TreeSitterTalon(TreeSitterTypeProvider):
    def resource_path(self, resource_name: str) -> Path:
        dirname = os.path.dirname(__file__)
        try:
            filename = Path(resource_filename("tree_sitter_talon", resource_name))
            if filename.exists() and filename.match(f"dirname/**"):
                return filename
        except (KeyError, ModuleNotFoundError):
            pass

        filename = Path(dirname) / resource_name
        if filename.exists():
            return filename

        raise FileNotFoundError(resource_name)

    @property
    def repository_path(self) -> str:
        return str(self.resource_path("data/tree-sitter-talon"))

    @property
    def data_path(self) -> Path:
        return self.resource_path("data")

    @property
    def library_path(self) -> str:
        library_name = {
            "linux": "talon.so",
            "darwin": "talon.dylib",
            "win32": "talon.dll",
        }[sys.platform]
        return str(self.data_path / library_name)

    @property
    def node_types_path(self) -> Path:
        return self.resource_path("data/tree-sitter-talon/src/node-types.json")

    def __init__(self, *, encoding: str = "utf-8"):
        # Read node-types.json
        node_types = NodeType.schema().loads( # type: ignore
            self.node_types_path.read_text(encoding=encoding), many=True
        )

        # Conversion from tree-sitter names to Python names
        def as_class_name(node_type_name: str) -> str:
            buffer = ["Talon"]
            for part in node_type_name.split("_"):
                buffer.append(part.capitalize())
            return "".join(buffer)

        # Initialize module
        super().__init__(
            "tree_sitter_talon",
            node_types,
            error_as_node=True,
            as_class_name=as_class_name,
            extra=["comment"],
        )

        # Build tree-sitter-talon
        tree_sitter.Language.build_library(self.library_path, [self.repository_path])
        self.language = tree_sitter.Language(self.library_path, "talon")
        self.parser = tree_sitter.Parser()
        self.parser.set_language(self.language)

        self.NodeTypeName = NodeTypeName
        self.NodeFieldName = NodeFieldName
        self.NodeTypeError = NodeTypeError
        self.Point = Point
        self.Node = Node
        self.Leaf = Leaf
        self.Branch = Branch

    def parse(
        self,
        contents: typing.Union[str, bytes],
        *,
        has_header: typing.Optional[bool] = None,
        encoding: str = "utf-8",
    ) -> typing.Union[typing.Sequence[Node], Node, None]:
        tree = self.parse_as_tree_sitter(
            contents, has_header=has_header, encoding=encoding
        )
        return self.from_tree_sitter(tree.root_node)

    def parse_file(
        self,
        path: typing.Union[str, Path],
        *,
        has_header: typing.Optional[bool] = None,
        encoding: str = "utf-8",
    ) -> typing.Union[typing.Sequence[Node], Node, None]:
        tree = self.parse_file_as_tree_sitter(
            path, has_header=has_header, encoding=encoding
        )
        return self.from_tree_sitter(tree.root_node)

    def parse_as_tree_sitter(
        self,
        contents: typing.Union[str, bytes],
        *,
        has_header: typing.Optional[bool] = None,
        encoding: str = "utf-8",
    ) -> tree_sitter.Tree:
        if isinstance(contents, str):
            contents = bytes(contents, encoding)
        if has_header is None:
            has_header = contents.startswith(b"-\n") or (b"\n-\n" in contents)
        if not has_header:
            contents = b"-\n" + contents
        return self.parser.parse(contents)

    def parse_file_as_tree_sitter(
        self,
        path: typing.Union[str, Path],
        *,
        has_header: typing.Optional[bool] = None,
        encoding: str = "utf-8",
    ) -> tree_sitter.Tree:
        if not isinstance(path, Path):
            path = Path(path)
        contents = path.read_bytes()
        return self.parse_as_tree_sitter(
            contents, has_header=has_header, encoding=encoding
        )


sys.modules[__name__] = TreeSitterTalon()



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from pathlib import Path
from typing import Optional, Union
from tree_sitter_type_provider import *

import tree_sitter # type: ignore

parser: tree_sitter.Parser

language: tree_sitter.Language


def parse(
    contents: Union[str, bytes],
    has_header: Optional[bool] = None,
    encoding: str = "utf-8",
) -> Node: ...


def parse_file(
    path: Union[str, Path],
    has_header: Optional[bool] = None,
    encoding: str = "utf-8",
) -> Node: ...


def parse_as_tree_sitter(
    contents: Union[str, bytes],
    has_header: Optional[bool] = None,
    encoding: str = "utf-8",
) -> tree_sitter.Tree: ...


def parse_file_as_tree_sitter(
    path: Union[str, Path],
    has_header: Optional[bool] = None,
    encoding: str = "utf-8",
) -> tree_sitter.Tree: ...


def from_tree_sitter(
    tsvalue: Union[tree_sitter.Tree, tree_sitter.Node, tree_sitter.TreeCursor],
    *,
    encoding: str = "utf-8"
) -> Node: ...


@dataclass_json
@dataclass
class TalonAction(Branch):
    children: list[TalonComment]
    action_name: TalonIdentifier
    arguments: TalonArgumentList


@dataclass_json
@dataclass
class TalonAnd(Branch):
    children: list[Union[TalonAnd, TalonMatch, TalonNot, TalonComment]]


@dataclass_json
@dataclass
class TalonArgumentList(Branch):
    children: list[Union[TalonAction, TalonBinaryOperator, TalonFloat, TalonInteger, TalonKeyAction, TalonParenthesizedExpression, TalonSleepAction, TalonString, TalonVariable, TalonComment]]


@dataclass_json
@dataclass
class TalonAssignment(Branch):
    children: list[TalonComment]
    left: TalonIdentifier
    right: Union[TalonAction, TalonBinaryOperator, TalonFloat, TalonInteger, TalonKeyAction, TalonParenthesizedExpression, TalonSleepAction, TalonString, TalonVariable, TalonComment]


@dataclass_json
@dataclass
class TalonBinaryOperator(Branch):
    children: list[TalonComment]
    left: Union[TalonAction, TalonBinaryOperator, TalonFloat, TalonInteger, TalonKeyAction, TalonParenthesizedExpression, TalonSleepAction, TalonString, TalonVariable, TalonComment]
    operator: TalonOperator
    right: Union[TalonAction, TalonBinaryOperator, TalonFloat, TalonInteger, TalonKeyAction, TalonParenthesizedExpression, TalonSleepAction, TalonString, TalonVariable, TalonComment]


@dataclass_json
@dataclass
class TalonBlock(Branch):
    children: list[Union[TalonAssignment, TalonDocstring, TalonExpression, TalonComment]]


@dataclass_json
@dataclass
class TalonCapture(Branch):
    children: list[TalonComment]
    capture_name: TalonIdentifier


@dataclass_json
@dataclass
class TalonChoice(Branch):
    children: list[Union[TalonCapture, TalonEndAnchor, TalonList, TalonOptional, TalonParenthesizedRule, TalonRepeat, TalonRepeat1, TalonSeq, TalonStartAnchor, TalonWord, TalonComment]]


@dataclass_json
@dataclass
class TalonCommand(Branch):
    children: list[TalonComment]
    rule: TalonRule
    script: TalonBlock


@dataclass_json
@dataclass
class TalonComment(Leaf):
    pass


@dataclass_json
@dataclass
class TalonContext(Branch):
    children: list[Union[TalonAnd, TalonDocstring, TalonMatch, TalonNot, TalonOr, TalonComment]]


@dataclass_json
@dataclass
class TalonDocstring(Leaf):
    pass


@dataclass_json
@dataclass
class TalonEndAnchor(Leaf):
    pass


@dataclass_json
@dataclass
class TalonError(Branch):
    children: list[Union[TalonAction, TalonAnd, TalonArgumentList, TalonAssignment, TalonBinaryOperator, TalonBlock, TalonCapture, TalonChoice, TalonCommand, TalonContext, TalonExpression, TalonIncludeTag, TalonInterpolation, TalonKeyAction, TalonList, TalonMatch, TalonNot, TalonNumber, TalonOptional, TalonOr, TalonParenthesizedExpression, TalonParenthesizedRule, TalonRegexEscapeSequence, TalonRepeat, TalonRepeat1, TalonRule, TalonSeq, TalonSettings, TalonSleepAction, TalonSourceFile, TalonString, TalonStringContent, TalonVariable, TalonComment, TalonDocstring, TalonEndAnchor, TalonFloat, TalonIdentifier, TalonImplicitString, TalonInteger, TalonOperator, TalonStartAnchor, TalonStringEscapeSequence, TalonWord, TalonError]]


@dataclass_json
@dataclass
class TalonExpression(Branch):
    children: list[TalonComment]
    expression: Union[TalonAction, TalonBinaryOperator, TalonFloat, TalonInteger, TalonKeyAction, TalonParenthesizedExpression, TalonSleepAction, TalonString, TalonVariable, TalonComment]


@dataclass_json
@dataclass
class TalonFloat(Leaf):
    pass


@dataclass_json
@dataclass
class TalonIdentifier(Leaf):
    pass


@dataclass_json
@dataclass
class TalonImplicitString(Leaf):
    pass


@dataclass_json
@dataclass
class TalonIncludeTag(Branch):
    children: list[TalonComment]
    tag: TalonIdentifier


@dataclass_json
@dataclass
class TalonInteger(Leaf):
    pass


@dataclass_json
@dataclass
class TalonInterpolation(Branch):
    children: list[Union[TalonAction, TalonBinaryOperator, TalonFloat, TalonInteger, TalonKeyAction, TalonParenthesizedExpression, TalonSleepAction, TalonString, TalonVariable, TalonComment]]


@dataclass_json
@dataclass
class TalonKeyAction(Branch):
    children: list[TalonComment]
    arguments: TalonImplicitString


@dataclass_json
@dataclass
class TalonList(Branch):
    children: list[TalonComment]
    list_name: TalonIdentifier


@dataclass_json
@dataclass
class TalonMatch(Branch):
    children: list[TalonComment]
    key: TalonIdentifier
    pattern: TalonImplicitString


@dataclass_json
@dataclass
class TalonNot(Branch):
    children: list[Union[TalonMatch, TalonComment]]


@dataclass_json
@dataclass
class TalonNumber(Branch):
    children: list[Union[TalonFloat, TalonInteger, TalonComment]]


@dataclass_json
@dataclass
class TalonOperator(Leaf):
    pass


@dataclass_json
@dataclass
class TalonOptional(Branch):
    children: list[Union[TalonCapture, TalonChoice, TalonEndAnchor, TalonList, TalonOptional, TalonParenthesizedRule, TalonRepeat, TalonRepeat1, TalonSeq, TalonStartAnchor, TalonWord, TalonComment]]


@dataclass_json
@dataclass
class TalonOr(Branch):
    children: list[Union[TalonAnd, TalonMatch, TalonNot, TalonComment]]


@dataclass_json
@dataclass
class TalonParenthesizedExpression(Branch):
    children: list[Union[TalonAction, TalonBinaryOperator, TalonFloat, TalonInteger, TalonKeyAction, TalonParenthesizedExpression, TalonSleepAction, TalonString, TalonVariable, TalonComment]]


@dataclass_json
@dataclass
class TalonParenthesizedRule(Branch):
    children: list[Union[TalonCapture, TalonChoice, TalonEndAnchor, TalonList, TalonOptional, TalonParenthesizedRule, TalonRepeat, TalonRepeat1, TalonSeq, TalonStartAnchor, TalonWord, TalonComment]]


@dataclass_json
@dataclass
class TalonRegexEscapeSequence(Branch):
    children: list[Union[TalonRegexEscapeSequence, TalonComment]]


@dataclass_json
@dataclass
class TalonRepeat(Branch):
    children: list[Union[TalonCapture, TalonList, TalonOptional, TalonParenthesizedRule, TalonRepeat, TalonRepeat1, TalonWord, TalonComment]]


@dataclass_json
@dataclass
class TalonRepeat1(Branch):
    children: list[Union[TalonCapture, TalonList, TalonOptional, TalonParenthesizedRule, TalonRepeat, TalonRepeat1, TalonWord, TalonComment]]


@dataclass_json
@dataclass
class TalonRule(Branch):
    children: list[Union[TalonCapture, TalonChoice, TalonEndAnchor, TalonList, TalonOptional, TalonParenthesizedRule, TalonRepeat, TalonRepeat1, TalonSeq, TalonStartAnchor, TalonWord, TalonComment]]


@dataclass_json
@dataclass
class TalonSeq(Branch):
    children: list[Union[TalonCapture, TalonList, TalonOptional, TalonParenthesizedRule, TalonRepeat, TalonRepeat1, TalonWord, TalonComment]]


@dataclass_json
@dataclass
class TalonSettings(Branch):
    children: list[Union[TalonBlock, TalonComment]]


@dataclass_json
@dataclass
class TalonSleepAction(Branch):
    children: list[TalonComment]
    arguments: TalonImplicitString


@dataclass_json
@dataclass
class TalonSourceFile(Branch):
    children: list[Union[TalonCommand, TalonContext, TalonIncludeTag, TalonSettings, TalonComment]]


@dataclass_json
@dataclass
class TalonStartAnchor(Leaf):
    pass


@dataclass_json
@dataclass
class TalonString(Branch):
    children: list[Union[TalonInterpolation, TalonStringContent, TalonStringEscapeSequence, TalonComment]]


@dataclass_json
@dataclass
class TalonStringContent(Leaf):
    pass


@dataclass_json
@dataclass
class TalonStringEscapeSequence(Leaf):
    pass


@dataclass_json
@dataclass
class TalonVariable(Branch):
    children: list[TalonComment]
    variable_name: TalonIdentifier


@dataclass_json
@dataclass
class TalonWord(Leaf):
    pass